import gym
from stable_baselines3 import A2C
from gymnasium.envs.registration import register
from tqdm import tqdm
import wandb
from wandb.integration.sb3 import WandbCallback
from stable_baselines3.common.vec_env import VecVideoRecorder
from huggingface_sb3 import push_to_hub


def train_model(config, env_id, policy, project_name):
    """
    Train a model using the A2C algorithm with Weights & Biases integration.

    Args:
        config (dict): Configuration parameters for training.
        env_id (str): Identifier of the Gym environment.
        policy (str): Type of policy to use for the model.
        project_name (str): Name of the project in Weights & Biases.

    Returns:
        A2C: Trained A2C model.
    """
    # Initialize Weights & Biases
    wandb.login()
    run = wandb.init(
        project=project_name,
        config=config,
        sync_tensorboard=True,  # auto-upload sb3's tensorboard metrics
        monitor_gym=True,  # auto-upload the videos of agents playing the game
        save_code=True,  # optional
    )

    # Register the environment
    register(id=env_id, entry_point='gym.envs.classic_control:CartPoleEnv', max_episode_steps=500)

    env = gym.make(env_id, render_mode="rgb_array")

    model = A2C(policy, env, verbose=1, tensorboard_log=f"runs/{run.id}")
    model.learn(total_timesteps=config["total_timesteps"])


    # Mark the run as public in W&B project settings
    run.finish()
    return model

def test_model(model):
    """
    Test a trained model by running it in the environment.

    Args:
        model (A2C): Trained A2C model to be tested.
    """
    vec_env = model.get_env()
    obs = vec_env.reset()

    for _ in tqdm(range(1000)):
        action, _state = model.predict(obs, deterministic=True)
        obs, reward, done, info = vec_env.step(action)
        vec_env.render("rgb_array")

def save_push_model(model, project_name): 
    """
    Save the trained model and push it to the Hugging Face Model Hub.

    Args:
        model (A2C): Trained A2C model.
        project_name (str): Name of the project to save the model.
    """
    model.save(project_name + ".zip")

    #HugingFace
    push_to_hub(
        repo_id="oscarchaufour/a2c-CartPole-v1",
        filename=project_name + ".zip",
        commit_message="Adding CartPole model trained with A2C on HuggingFace",
        token="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"
    )


if __name__ == "__main__":
    env_id = "CartPole-v1"
    policy = "MlpPolicy"

    config = {
    "policy_type": policy,
    "total_timesteps": 10000,
    "env_name": env_id,
    }
    project_name = "a2c-CartPole-v1"

    trained_model = train_model(config, env_id, policy, project_name)
    test_model(trained_model)
    save_push_model(trained_model, project_name)
