import gym
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from tqdm import tqdm 
import numpy as np
from torch.distributions import Categorical

# Define the neural network model
class Policy(nn.Module):
    def __init__(self, input_size, output_size, hidden_size=128):
        super(Policy, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.5)
        self.fc2 = nn.Linear(hidden_size, output_size)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.fc1(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.softmax(x)
        return x

def reinforce(policy, env, optimizer):
    # Training loop
    num_episodes = 200
    gamma = 0.99
    episodes_rewards = []

    for episode in tqdm(range(num_episodes)):
        # Reset the environment and get the initial observation
        observation = env.reset()

        rewards = []
        log_probabilities = []
        terminated = False
        episode_reward = 0

        while not terminated:
            # Compute action probabilities
            observation_array = observation[0] if isinstance(observation, tuple) else observation
            observation_tensor = torch.FloatTensor(observation_array).unsqueeze(0)
            action_probabilities = policy(observation_tensor)
            
            cat = Categorical(action_probabilities)
            # Sample action based on probabilities and store its probability in the buffer
            action = cat.sample()
            log_probability = cat.log_prob(action)

            # Step the environment with the action
            observation, reward, terminated, _, _ = env.step(action.item())
            env.render()

            # Compute and store the reward in the buffer
            episode_reward += reward
            log_probabilities.append(log_probability)
            rewards.append(reward)
        
        episodes_rewards.append(episode_reward)

        # Compute the returns
        weighted_rewards = 0
        returns = []
        for r in rewards[::-1]:
            weighted_rewards = r + gamma * weighted_rewards
            returns.insert(0, weighted_rewards)
        returns = torch.tensor(returns)
        returns = (returns - returns.mean()) / (returns.std() + 1e-5)

        # Compute policy loss
        loss = []
        for log_prob, weighted_rewards in zip(log_probabilities, returns):
            loss.append(log_prob * weighted_rewards)
        policy_loss = - torch.cat(loss).sum()

        # Update the policy
        optimizer.zero_grad()
        policy_loss.backward()
        optimizer.step()
    return episodes_rewards


def plot_rewards(episodes_rewards):
    # Plot the total reward across episodes
    plt.plot(episodes_rewards)
    plt.xlabel('Episode')
    plt.ylabel('Total Reward')
    plt.title('REINFORCE: Total Reward across Episodes')
    plt.show()


if __name__ == "__main__":
    # Create the environment
    env = gym.make("CartPole-v1", render_mode="human")

    # Set up the agent
    policy = Policy(
        input_size=env.observation_space.shape[0],
        output_size=env.action_space.n
    )
    optimizer = optim.Adam(policy.parameters(), lr=5e-3)
    episodes_rewards = reinforce(policy, env, optimizer)
    env.close()

    # plot the rewards
    plot_rewards(episodes_rewards)


    # Serialize the model and save it to a .zip file
    # import pickle
    # import zipfile

    # Assuming `policy` is your neural network model

    # # Step 1: Serialize the model
    # model_bytes = pickle.dumps(policy)

    # # Step 2: Create a .zip file containing the serialized model
    # zip_filename = ".zip"
    # with zipfile.ZipFile(zip_filename, 'w') as zipf:
    #     zipf.writestr("model.pkl", model_bytes)

